/*
   Servo Controller. Atmel ATmega48/88/168 based.
    INT Version (v1.2)

  ,  EEPROM  UART.
 - ""   ( / LoadPIDParamRAM).
 
   -  ServoTuning   Servo_int_AtMega88.hex
  ServoUART.hex     hex (2),   
 / LoadPIDParamRAM,   ""  . 

   Clock: 16 MHz


        
   :

  BOOTSZ1 = 0 (  )
  BOOTSZ0 = 0
  BOOTRST = 1 (   )
  RSTDISBL = 1 ( Reset:0-In/Out; 1-Reset) !!!

  DWEN = 1  (0 - , 1 -  debugWire)
  WDTON = 1 ( :0- , 1-  )
  EESAVE = 1 ( "  EEPROM: 0- ,1-)

  BODLEVEL2 = 1 (  
  *BODLEVEL1 = 0    4,3)
  *BODLEVEL0 = 0

  *CKDIV8 = 1	  8   . 0-, 1- . 
  CKOUT = 1 	        

  SUT1 = 1		 
  SUT0 = 0		 
  *CKSEL3 = 1 	   
  *CKSEL2 = 1    o    
  CKSEL1 = 1	 F = 16 
  CKSEL0 = 0 	 
*/

;  ; (  )        
;.include "m48def.inc"	;  AtMega48   / SavePIDParamEEPROM 
						;  LoadPIDParamEEPROM ""  1 				

.include "m88def.inc"

;-------------------------------------
; PORT B
.equ	B_PWM_OUT1 = 1		;   1
.equ	B_PWM_OUT2 = 2		;   2


; PORT C
.equ	C_SWITCH2 = 2		;   2
.equ	C_DIRECTION = 3		;  , 

; PORT D
.equ	D_NPC_STEP = 2  	; INT0,   ()
;.equ	D_RESERVED = 3		; INT1, reserved
.equ	D_ENCODER_B = 5 	; encoder, B-input
.equ	D_SWITCH1 = 7		;   1


;-------------------------------------

.def	_Flags		= r23
.def	uart_r		= r24 	;  ,    UART_OUT

.def	_0			= r15 	; 

.def	_PvEnc		= r14 	;    
.def	_PvDir		= r13 	;    

.def	_PosH		= r12 	;	24-bit    
.def	_PosM		= r11 	;
.def	_PosL		= r10 	;

.def	_200HzCntL	= r4  	;  200 Hz 
.def	_200HzCntH	= r5  	;

.def	_VelocityL	= r6  	;    
.def	_VelocityH	= r9  	;

.def	_CtDiv83	= r7  	;  1/83 (1kHz)
.def	_CtDiv4		= r25 	;  1/5 (200 Hz)
.def	_Mode		= r8  	;  -	


;-------------------------------------
; _Flags
;-------------------------------------
.equ	full_bit		=0
.equ	command_bit		=1 ;      

.equ	wait_cmd_bit	=2	;  
.equ	wait_size_bit	=3	;  

.equ	_1khz_bit		=4 ; 1 
.equ	TorqLim_bit		=5 ;  
	
.equ	pwma_bit		=6	; 
.equ	pwmb_bit		=7	; 
;-------------------------------------




.dseg

		.org	0x100

VelError: 		.byte  2	;  	
VelError200: 	.byte  2	;   (200Hz)
Velocity: 		.byte  2 	; 	

CtPos:	.byte	3			; 24-bit  		g/j	mode 3
CtSub:	.byte	2			; Sub command   		s	mode 0/1/2
PvInt:	.byte	2			; Integration register
PvPos:	.byte	2			; 16-bit Velocity detection register

Parms:
LimSpd:	.byte	2			;P0,Velocity limit		Integer
GaSpd:	.byte	2			;P1,Velocity feedback gain	Fixed point
GaTqP:	.byte	2			;P2,Proportional gain		Fixed point
GaTqI:	.byte	2			;P3,Integral gain		Fixed point
LimTrq:	.byte	2			;P4,Torque limit		Integer
GaEG:	.byte	2			;P5,EG feedback gain		Fixed point



UART_COMMAND: 	.byte 1
UART_DATA_SIZE:	.byte 1

UART_OUT_T: 	.byte 1
UART_OUT_H:		.byte 1

UART_IN:		.byte 	16

UART_OUT:		.byte	16


;----------------------------------------------------------;
; EEPROM
;----------------------------------------------------------;
.eseg
	; Memory bank 0
	.dw	300, 0x0F00, 0x0300, 0x0060, 180, 0x0550 ;,  0x1a00, 32

	; Memory bank 1 : Tamagawa Seiki TS1410N1
	.dw	300, 0x0800, 0x0300, 0x0060, 180, 0x0550 ;,  0x1a00, 32

	; Memory bank 2 : Yasukawa Electric UGTMEM-A1SA51
	.dw	300, 0x0500, 0x0180, 0x0060, 255, 0x0340 ;,  0x2600, 128

	; Memory bank 3 : Matsushita Electric MCN-14EAEC (6V, 40p/r)
	.dw	300, 0x1E00, 0x0400, 0x0019, 230, 0x0840 ;,  0x1400, 64

	; Memory bank 4 :  
	.dw	300, 0x09C4, 0x0640, 0x0060, 180, 0x0550

	.org	$0100
;	.db	 (1<<TOIE0), (1<<INT0), DEF_SERVO_CYCLE, DEF_SERVO_CYCLE_SOFT, QUARTZ_FREQ
	; QUARTZ_FREQ -   ,   


.equ	N_PARM = 6;8	; Number of parameters par bank.
;----------------------------------------------------------;


.cseg
		.org	$0000

	RJMP 	Start     		;   reset
							;     
							; (- )    
	RJMP 	INT0_INT		; External Interrupt Request 0
	RJMP 	INT1_INT		; External Interrupt Request 1
	RJMP 	Start			; PCINT0		; Pin Change Interrupt Request 0
	RJMP 	Start			; PCINT1		; Pin Change Interrupt Request 0
	RJMP 	Start			; PCINT2		; Pin Change Interrupt Request 1
	RJMP 	Start			; WDT			; Watchdog Time-out Interrupt
	RJMP 	Start			; TIM2_COMPA	; Timer/Counter2 Compare Match A
	RJMP 	Start			; TIM2_COMPB	; Timer/Counter2 Compare Match B
	RJMP 	TIMER2_OVF_INT	; Timer/Counter2 Overflow
	RJMP 	Start			; TIM1_CAPT		; Timer/Counter1 Capture Event
	RJMP 	Start			; TIM1_COMPA	; Timer/Counter1 Compare Match A
	RJMP 	Start			; TIM1_COMPB	; Timer/Counter1 Compare Match B
	RJMP 	Start			; TIM1_OVF		; Timer/Counter1 Overflow
	RJMP 	Start			; TIM0_COMPA	; TimerCounter0 Compare Match A
	RJMP 	Start			; TIM0_COMPB	; TimerCounter0 Compare Match B
	RJMP 	TIMER0_OVF_INT	; TIM_OVF	 	; Timer/Couner0 Overflow
	RJMP 	Start			; SPI_SRC		; SPI Serial Transfer Complete
	RJMP 	Start			; USART Rx Complete
	RJMP 	Start			; USART, Data Register Empty
	RJMP 	Start			; USART_TXC		; USART Tx Complete
	RJMP	Start			; ADC Conversion Complete
	RJMP 	Start			; EE_RDY		; EEPROM Ready
	RJMP 	Start			; ANA_COMP		; Analog Comparator
	RJMP 	Start			; TWI			; Two-wire Serial Interface
	RJMP 	Start			; SPM_RDY		; Store Program Memory Read

	
Start:
		clr		_0
		ldi		r16, low(RAMEND)
		out		spl, r16
		ldi		r16, HIGH(RAMEND)
		out		sph, r16
;-------------------------------------------------------
		wdr								;   
		lds 	r16,WDTCSR				;
		ori		r16,(1<<WDCE)|(1<<WDE)	;	
		sts 	WDTCSR,r16				;  WDT 
		ldi		r16,(1<<WDE)|(1<<WDP0)	;   
		sts 	WDTCSR,r16				; t=32ms
;-----------------------------------------------------------

		;  RAM
		ldi		YL, LOW(0x100)
		ldi		YH, HIGH(0x100)

		ldi		r16, 0 ;
		ldi		r17, 4 ; 1024 bytes
cl_ram:	st		Y+, _0
		sec
		sbc		r16, _0
		sbc		r17, _0
		cp		r16, _0
		cpc		r17, _0
		brne	cl_ram
		;-------

; ----------------------------(  (pin)  ATMega88   TQPF-32)
;**********   B: **********************************************
	ldi	r16,0b11111111			; PB7(pin8) - XTAL2
	out	DDRB,r16				; PB6(pin7) - XTAL1
	ldi	r16,0b00000000			; PB5(pin17) - SCK
	out	PORTB,r16				; PB4(pin16) - MISO
								; PB3(pin15) - MOSI
;     	; PB2(pin14) -  1 (-2)
;     			; PB1(pin13) -  1 (-1)
;  ""  0				; PB0(pin12) -  ( .)
	
;**********   : **********************************************
	ldi	r16,0b11110110			; PC7(xxx) -    
	out	DDRC,r16				; PC6(pin29) - Reset ( .)
	ldi	r16,0b00000000			; PC5(pin28) -  ( .) 
	out	PORTC,r16				; PC4(pin27) -  ( .)
								; PC3(pin26) -  (DIRECT (F/R) - ) 
							  	; PC2(pin25) -  ( .)
								; PC1(pin24) -  ( .)
								; PC0(pin23) -  ( .)
	
;**********   D: **********************************************
	ldi	r16,0b11001010			; PD7(pin11) -  ( .)
	out	DDRD,r16				; PD6(pin10) -  ( .)
	ldi	r16,0b00000000			; PD5(pin9) -    
	out	PORTD,r16				; PD4(pin2) -    
								; PD3(pin1) -  ( .)
;     	; PD2(pin32) -  Step 
;     			; PD1(pin31) -  TX 
; ""  0				; PD0(pin30) -  RX 

;----------------------------------------------------------------
		ldi	r16,0b10000000		;  .
		out	ACSR,r16			;
;----------------------------------------------------------------
; timer 0 -  , -
		ldi		r16,0b00000011; CK/64 prescaler	
		out		TCCR0B, r16	; (976)
;---------------------------------------------------------------
; timer1 -  , (Phase correct PWM, 8-bit, )
		ldi		r16,0b10100001	;
		sts		TCCR1A, r16		;

		ldi		r16,0b00000001	;   
		sts		TCCR1B, r16
		
		sts		OCR1AH, _0
		sts		OCR1AL, _0
		sts		OCR1BH, _0
		sts		OCR1BL, _0
;-------------------------------------------------------------------
; timer 2 - - (  )
		ldi		r16,0b00000111	; CK/1024 prescaler	
		sts		TCCR2B, r16
;--------------------------------------------------------------------
		ldi		r16,0b00000001	;   
		sts		TIMSK0,r16		;    Timer0 
		sts		TIMSK2,r16		; & Timer2
;---------------------------------------------------------------------
		ldi		r16,0b00000010	;INT0    (rising ISC01 | ISC00)
		sts		EICRA, r16		;
	
		ldi		r16,0b00000001	;   INT0 (Step)
		out		EIMSK, r16		;
;-----------------------------------------------------------------

		;  3- -
		ldi		r16, 3
		mov		_Mode, r16
		rcall	InitServo

		;   PID-  RAM
		rcall	LoadPIDParamRAM

		ldi		r16,0b10011000
		sts		UCSR0B, r16
		sei						;   
	


Main:
		nop							;
		rjmp	main				;



;================================================================================
;  
;================================================================================
InitServo:
		ldi		yl, low(CtPos)
		ldi		yh, high(CtPos)
		ldi		r16, 9
cl_pos:	st		Y+, _0
		dec		r16
		brne	cl_pos
		clr		_PosL
		clr		_PosM
		clr		_PosH
 		ret

;================================================================================
;================================================================================
;   PID-    ( EEPROM)
; ===============================================================================
;       ServoTuning  
;   hex- ( 2 ).   (Integer),    hex, 
;    (Fixed point)     ServoTuning  
;  256     ->  hex (2 )
;================================================================================
LoadPIDParamRAM:
		ldi		yl, low(Parms)
		ldi		yh, high(Parms)
;----------------------------------------------------------------------			
; LimSpd -  P0,Velocity limit	( )
; 300 -  
	ldi		r16,0x2C			; .
	st		Y+, r16				;	
	ldi		r16,0x01			; .
	st		Y+, r16				;	
;----------------------------------------------------------------------
;GaSpd -  P1,Velocity feedback gain	( ) 
; 30 - ,      ( 256)
; (    (K F))  
	ldi		r16,0x00			; .
	st		Y+, r16				;	
	ldi		r16,0x1E			; .
	st		Y+, r16				;	
;----------------------------------------------------------------------
; GaTqP - P2,Proportional gain	( )
; 4  -   (.)  ( 256)
;      .   
;    .     
;   
	ldi		r16,0x00			; .
	st		Y+, r16				;	
	ldi		r16,0x04			; .
	st		Y+, r16				;	
;----------------------------------------------------------------------
; GaTqI - P3,Integral gain	( )
; 0,1 -   (.)    ( 256)
	ldi		r16,0x19			; .
	st		Y+, r16				;	
	ldi		r16,0x00			; .
	st		Y+, r16				;	
;----------------------------------------------------------------------
; LimTrq - P4,Torque limit		( )
; 250 -    
	ldi		r16,0xFA			; .
	st		Y+, r16				;	
	ldi		r16,0x00			; .
	st		Y+, r16				;	
;----------------------------------------------------------------------
; GaEG -  P5,EG feedback gain	( )     ( 256)
; 1360 -    ( 
;    ())
	ldi		r16,0x50			; .
	st		Y+, r16				;	
	ldi		r16,0x05			; .
	st		Y+, r16				;	
;----------------------------------------------------------------------
	ret
;===========================================================================


;================================================================================
;   INT1 - 
;================================================================================
INT1_INT:
		push	r0
		in		r0, SREG

		sec

;		sbis	PIND, D_RESERVED
;		rjmp	i1_fall	

; rising edge
		sbis	PIND, D_ENCODER_B
		rjmp	i1sub
		rjmp	i1add
; falling edge
;i1_fall:
;		sbic	PIND, D_ENCODER_B
;		rjmp	i1sub

i1add:  cli
		adc		_PosL, _0
		adc		_PosM, _0
		adc		_PosH, _0
		sei
		rjmp	i1out

i1sub:  cli
		sbc		_PosL, _0
		sbc		_PosM, _0
		sbc		_PosH, _0
		sei

i1out:	out		SREG, r0
		pop		r0
		reti


;================================================================================
;   INT0 -    
;================================================================================
INT0_INT:
		push	r0
		in		r0, SREG		 ; Save carry flag
		push	zl

		sec
		lds		zl,	CtPos+0
		sbic	PINC, C_DIRECTION ; F/R = 0?	 
		rjmp	i0frw		      ; 
 
		; 
		sbc		zl, _0
		sts		CtPos+0, zl
		lds		zl, CtPos+1
		sbc		zl, _0
		sts		CtPos+1, zl
		lds		zl, CtPos+2
		sbc		zl, _0
		rjmp	i0so

i0frw:	; 
		adc		zl, _0
		sts		CtPos+0, zl
		lds		zl, CtPos+1
		adc		zl, _0
		sts		CtPos+1, zl
		lds		zl, CtPos+2
		adc		zl, _0

i0so:	sts		CtPos+2, zl
		pop		zl
		out		SREG, r0
		pop		r0
		reti


;================================================================================
;    TIMER2 - 16ms=62,5Hz
;================================================================================
TIMER2_OVF_INT:
		push	r0
		push	zl
		push	zh
		in		r0, SREG
		
;		lds		zl, ServoCycleTimeByte
		ldi		zl,0b00000000	;
		sts		TCNT2, zl		; clock divisor

		wdr							;   
		
		rjmp	tc2_Xkhz

tc2_eoi:
		out		SREG, r0
		pop		zh
		pop		zl
		pop		r0
		reti	


;================================================================================
;    TIMER0 - 83.3 
;================================================================================
TIMER0_OVF_INT:
		push	r0
		push	zl
		push	zh
		in		r0, SREG

		ldi		zl, -3 			; b11111101 (253, $FD)
		out		TCNT0, zl		; Restore Clock divisor


		mov		zl, _PvEnc		;   zl[1:0]
		
		in		_PvEnc, PIND	;     _PvEnc[5:4]
		swap	_PvEnc			;   _PvEnc[1:0]

		ldi		zh, 1

		sbrc	_PvEnc, 1		; Skip,    A  
		eor		_PvEnc, zh		; B xor 1

		; A B  A B
		; 1 1  1 0
		; 0 0  0 0
		; 1 0  1 1
		; 0 1  0 1

		sub		zl, _PvEnc		; Decode motion (zl = zl - _PvEnc)
		andi	zl, 3			; zl = zl and b11
		breq	enc_zr			; 00 -  
		cpi		zl, 1
		breq	enc_inc			; 01 -  
		cpi		zl, 3
		breq	enc_dec			; 11 -  

		; 10 - 
		mov		zl, _PvDir
		mov		zh, _PvDir		; Count by 2 for previous direction
		lsl		zl				; zl=zl shl 1, bit0 cleared
		asr		zh				; zh=zh shr 1, bit7=const, bit0 -> carry flag	
		rjmp	enc_add

enc_dec:
		ldi		zl, low(-1)
		ldi		zh, high(-1)
		rjmp	enc_setpvdir

enc_inc:
		ldi		zl, low(1)
		ldi		zh, high(1)
enc_setpvdir:
		mov		_PvDir, zl
enc_add:
		add		_PosL, zl		;  
		adc		_PosM, zh
		adc		_PosH, zh
enc_zr:
		dec		_CtDiv83		;  1/83 
		breq	tc0_1khz		;  0

tc0_eoi:
		out		SREG, r0
		pop		zh
		pop		zl
		pop		r0
		reti



;************************************************8
; SERVO, 1kHz
;************************************************8
tc0_1khz:
		ldi		zl, 83 ; (16MHz- 83, 110592 - 57.6, 10 MHz - 52; 8MHz - 42)
		mov		_CtDiv83, zl		; Re-initialize 1/83 divider

tc2_XkHz:
		sei
		push	r0
		push	r1
		push	r2
		push	r3
		push	r16
		push	r17
		push	r18
		push	r19
		push	r20
		push	r21
		push	r22

		;
		;
		; 200 HZ
		;
		dec		_CtDiv4			;  1/4 
		brne	no_200hz		;  0

		;   200Hz
		sec
		adc		_200HzCntL, _0
		adc		_200HzCntH, _0


		sts		Velocity+0, _VelocityL
		sts		Velocity+1, _VelocityH
		
		lds		zl, VelError+0
		sts		VelError200+0, zl
		lds		zl, VelError+1
		sts		VelError200+1, zl
		
		sts		VelError, _0
		sts		VelError+1, _0
		clr		_VelocityL
		clr		_VelocityH
		
		ldi		_CtDiv4, 5

no_200hz:
		lds		r2, PvPos+0		;  
		lds		r3, PvPos+1
		cli		
		sub		r2, _PosL
		sbc		r3, _PosM
		sts		PvPos+0, _PosL
		sts		PvPos+1, _PosM
		sei

		;   
		add		_VelocityL, r2
		adc		_VelocityH, r3

		com		r2				; $ff - r2
		com		r3				; $ff - r3
		adc		r2, _0
		adc		r3, _0			; r3:r2 = Velocity

		;  
		mov		r16, _Mode
		cpi		r16, 3			;  ( 3)
		breq	tap_position

		lds		r0, CtSub+0
		lds		r1, CtSub+1		; S- ( 0/1/2)
		mov		zl, r0
		mov		zh, r1
		mov		r19, r0
		mov		r20, r1
		cpi		r16, 2			;  ( 2)
		breq	tap_velocity
		cpi		r16, 1
		brne	PC+2
		rjmp	tap_torque		;   ( 1)
		rjmp	tap_voltage		;  ( 0)

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_position:
		cli
		lds		r16, CtPos+0	; 24-  
		lds		r17, CtPos+1
		lds		r18, CtPos+2
		
		sub		r16, _PosL		;   (dX)
		sbc		r17, _PosM
		sbc		r18, _PosH
		sei

		lds		r19, LimSpd+0   ;  
		lds		r20, LimSpd+1
		clr		r21
		cp		r16, r19		; 
		cpc		r17, r20
		cpc		r18, r21
		brge	pVovr; PC+11

		com		r19				; $ff - r19
		com		r20				; $ff - r20
		com		r21				; $ff - r21 = $ff

		adc		r19, _0
		adc		r20, _0
		adc		r21, _0

		cp		r16, r19
		cpc		r17, r20
		cpc		r18, r21
		brge	pVovr2 ; PC+3

pVovr:	mov		r16, r19
		mov		r17, r20
pVovr2:	mov		r0, r16
		mov		r1, r17			; r1:r0 = velocity command


;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
; r3:r2
; r1:r0 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_velocity:
		mov		r16, r2
		mov		r17, r3			; Velocity loop gain ( )

		lds		r18, GaSpd+0
		lds		r19, GaSpd+1
		rcall	muls1616		; r20:19 - scaled velocity

		sub		r0, r19			;  
		sbc		r1, r20

		;   -  
		lds		r16, VelError+0
		lds		r17, VelError+1
		add		r16, r0
		adc		r17, r1
		sts		VelError+0, r16	;  
		sts		VelError+1, r17	;  
		

		mov		r16, r0			;  
		mov		r17, r1			;  	
		lds		r18, GaTqP+0	;
		lds		r19, GaTqP+1	;
		rcall	muls1616		; r20:19 - proportional term
		mov		zl, r19
		mov		zh, r20
		
		lds		r16, PvInt+0	;  
		lds		r17, PvInt+1	;  	
		add		r0, r16			;
		adc		r1, r17			;
		mov		r16, r0			;
		mov		r17, r1			;
		lds		r18, GaTqI+0	;
		lds		r19, GaTqI+1	;
		rcall	muls1616		; r20:r19 - integral term
		add		zl, r19			;
		adc		zh, r20			; z:=proportional term + integral term

		lds		r18, LimTrq+0	;  
		ldi		r19, 0			;
		cp		zl, r18			;
		cpc		zh, r19			;
		brge	PC+6			;
		neg		r18				; $00 - r18
		com		r19				; $ff - r19
		cp		zl, r18			;
		cpc		zh, r19			;
		brge	PC+5			;
		mov		zl, r18			;
		mov		zh, r19			;
		sbr		_Flags, (1<<TorqLim_bit)	;
		rjmp	PC+6			;---
		sts		PvInt+0, r0		; Anti-wind-up
		sts		PvInt+1, r1		;
		cbr		_Flags, (1<<TorqLim_bit)	;


;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_torque:
		mov		r16, r2
		mov		r17, r3			; EG compensation
		lds		r18, GaEG+0		;
		lds		r19, GaEG+1		;
		rcall	muls1616		;
		add		r19, zl			;
		adc		r20, zh			; r20:r19 - Voltage command

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
; r20:r19
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_voltage:
		ldi		r16, low(255)	;
		ldi		r17, high(255)	;   -255  +255.
		cp		r19, r16		;
		cpc		r20, r17		;
		brge	t6 ;PC+6		;

		ldi		r16, low(-255)	;
		ldi		r17, high(-255)	;
		cp		r19, r16		;
		cpc		r20, r17		;
		brge	mo ;PC+3		;

t6:		mov		r19, r16		;
		mov		r20, r17		;/ r20:r19


mo:		tst		r19				; ??
		brne	tdir			;
		cbi		PORTD, D_SWITCH1
		cbi		PORTC, C_SWITCH2
		sts		OCR1AH, _0
		sts		OCR1AL, _0
		sts		OCR1BH, _0
		sts		OCR1BL, _0
		rjmp	ep

tdir:
		tst		r20				;Motor output. Set PWM register and polarity
		brmi	sw1   	;PC+5	;
	
		; ONE DIR
		cbi		PORTC, C_SWITCH2
		cbi		PORTB, B_PWM_OUT2
		sts		OCR1BL, _0
		sts		OCR1AL, r19
		sbi		PORTD, D_SWITCH1
		rjmp	ep 		; PC+5	

sw1:
		; ANOTHER DIR
		cbi		PORTD, D_SWITCH1
		cbi		PORTB, B_PWM_OUT1
		neg		r19
		sts		OCR1AL, _0
		sts		OCR1BL, r19
		sbi		PORTC, C_SWITCH2

/*;---------------------------------------------------------------------------
mo: 	
	tst r19 						;   cbcb128 ( cnczone.com) 
	brne 	tdir 					;   Dead-Time  
	cbi 	PORTD, D_SWITCH1		; 
	cbi 	PORTC, C_SWITCH2
	cbr 	_Flags, (1<<pwma_bit) 	;Add to die area time

	cbr 	_Flags, (1<<pwmb_bit) 	;Close the designation of PWM 
	sts 	OCR1AL, _0
	sts 	OCR1AH, _0
	sts 	OCR1BL, _0
	sts 	OCR1BH, _0
	rjmp 	ep

tdir:
	tst 	r20 					;Motor output. Set PWM register and polarity
	brmi 	sw1 					;PC+5 ;

; ONE DIR
	cbi 	PORTC,C_SWITCH2
	cbi 	PORTB,B_PWM_OUT2
	sts 	OCR1BL, _0
	sbrs 	_Flags, pwma_bit 		;If the PWMA indicate to 1 jumps to turn
	rcall 	Delay1us 				;Delay1US 

	sts 	OCR1AL, r19
	sbi 	PORTD, D_SWITCH1
	sbr 	_Flags, (1<<pwma_bit) 	;??PWMA??
	cbr 	_Flags, (1<<pwmb_bit) 	;??PWMB??
	rjmp 	ep 						; PC+5 

sw1:
		; ANOTHER DIR
	cbi 	PORTD, D_SWITCH1
	cbi 	PORTB, B_PWM_OUT1
	neg 	r19
	sts 	OCR1AL, _0
	sbrs 	_Flags, pwmb_bit
	rcall 	Delay1us
	sts 	OCR1BL, r19
	sbi 	PORTC, C_SWITCH2
	sbr 	_Flags, (1<<pwmb_bit)
	cbr 	_Flags, (1<<pwma_bit)
;-----------------------------------------------------------------
*/

ep:		pop		r22
		pop		r21
		pop		r20
		pop		r19
		pop		r18
		pop		r17
		pop		r16
		pop		r3
		pop		r2
		pop		r1
		pop		r0
		sbr		_Flags, (1<<_1khz_bit)	; 1kHz interrupt
		rjmp	tc0_eoi


Delay1us:
;	nop
	rjmp	PC+1
	ret



; == END SERVO 1kHz ================



;================================================================================
;  16*16 (signed)
; : 32- (r21:r20:r19:r18)
; : 16- (r19:r18)
; : 16- (r17:r16)
;================================================================================
muls1616:
		clt
		tst		r17
		brpl	PC+6
		set
		com		r16
		com		r17
		adc		r16,_0
		adc		r17,_0

		sub		r20,r20	; clear high 16bit.
		sub		r21,r21	;
		ldi		r22,17	; r22 = loop count
		brcc	PC+3	; ---- calcurating loop
		add		r20,r16	;
		adc		r21,r17	;
		ror		r21	;
		ror		r20	;
		ror		r19	;
		ror		r18	;
		dec		r22	; if (--r22 > 0)
		brne	PC-8	;  continue loop;

		brtc	mulret ;PC+9	; Negate the result if multiplyer was negative.
		com		r18
		com		r19
		com		r20
		com		r21
		adc		r18, _0
		adc		r19, _0
		adc		r20, _0
		adc		r21, _0
mulret:	ret



















