/*
   Servo Controller. Atmel ATmega48/88/168 based.
    INT Version (v1.0)

   Clock: 16 MHz


        
   :

  BOOTSZ1 = 0 (  )
  BOOTSZ0 = 0
  BOOTRST = 1 (   )
  RSTDISBL = 1 ( Reset:0-In/Out; 1-Reset) !!!

  DWEN = 1  (0 - , 1 -  debugWire)
  WDTON = 1 ( :0- , 1-  )
  EESAVE = 1 ( "  EEPROM: 0- ,1-)

  BODLEVEL2 = 1 (  
  *BODLEVEL1 = 0    4,3)
  *BODLEVEL0 = 0

  *CKDIV8 = 1	  8   . 0-, 1- . 
  CKOUT = 1 	        

  SUT1 = 1		
  SUT0 = 0		 
  *CKSEL3 = 1 	   
  *CKSEL2 = 1  	 o    
  CKSEL1 = 1	 F = 16 
  CKSEL0 = 0 	 
*/

;  ; (  )        
;.include "m48def.inc"	;  AtMega48   / SavePIDParamEEPROM 
						;  LoadPIDParamEEPROM ""  1 				

.include "m88def.inc"
;-------------------------------------
; PORT B
.equ	B_PWM_OUT1 = 1		;   1
.equ	B_PWM_OUT2 = 2		;   2


; PORT C
.equ	C_SWITCH2 = 2		;   2
.equ	C_DIRECTION = 3		;  , 

; PORT D
.equ	D_NPC_STEP = 2  	; INT0,   ()
;.equ	D_RESERVED = 3		; INT1, reserved
.equ	D_ENCODER_B = 5 	; encoder, B-input
.equ	D_SWITCH1 = 7		;   1


;-------------------------------------

.def	_Flags		= r23
.def	uart_r		= r24 	;  ,    UART_OUT

.def	_0			= r15 	; 

.def	_PvEnc		= r14 	;    
.def	_PvDir		= r13 	;    

.def	_PosH		= r12 	;	24-bit    
.def	_PosM		= r11 	;
.def	_PosL		= r10 	;

.def	_200HzCntL	= r4  	;  200 Hz 
.def	_200HzCntH	= r5  	;

.def	_VelocityL	= r6  	;    
.def	_VelocityH	= r9  	;

.def	_CtDiv83	= r7  	;  1/83 (1kHz)
.def	_CtDiv4		= r25 	;  1/5 (200 Hz)
.def	_Mode		= r8  	;  -	


;-------------------------------------
; _Flags
;-------------------------------------
.equ	full_bit		=0
.equ	command_bit		=1 ;      

.equ	wait_cmd_bit	=2	;  
.equ	wait_size_bit	=3	;  

.equ	_1khz_bit		=4 ; 1 
.equ	TorqLim_bit		=5 ;  
	
.equ	pwma_bit		=6	; 
.equ	pwmb_bit		=7	; 

;-------------------------------------




.dseg

		.org	0x100

VelError: 		.byte  2	;  	
VelError200: 	.byte  2	;   (200Hz)
Velocity: 		.byte  2 	; 	

CtPos:	.byte	3			; 24-bit  		g/j	mode 3
CtSub:	.byte	2			; Sub command   		s	mode 0/1/2
PvInt:	.byte	2			; Integration register
PvPos:	.byte	2			; 16-bit Velocity detection register

Parms:
LimSpd:	.byte	2			;P0,Velocity limit		Integer
GaSpd:	.byte	2			;P1,Velocity feedback gain	Fixed point
GaTqP:	.byte	2			;P2,Proportional gain		Fixed point
GaTqI:	.byte	2			;P3,Integral gain		Fixed point
LimTrq:	.byte	2			;P4,Torque limit		Integer
GaEG:	.byte	2			;P5,EG feedback gain		Fixed point



UART_COMMAND: 	.byte 1
UART_DATA_SIZE:	.byte 1

UART_OUT_T: 	.byte 1
UART_OUT_H:		.byte 1

UART_IN:		.byte 	16

UART_OUT:		.byte	16


;----------------------------------------------------------;
; EEPROM
;----------------------------------------------------------;
.eseg
	; Memory bank 0
	.dw	300, 0x0F00, 0x0300, 0x0060, 180, 0x0550 ;,  0x1a00, 32

	; Memory bank 1 : Tamagawa Seiki TS1410N1
	.dw	300, 0x0800, 0x0300, 0x0060, 180, 0x0550 ;,  0x1a00, 32

	; Memory bank 2 : Yasukawa Electric UGTMEM-A1SA51
	.dw	300, 0x0500, 0x0180, 0x0060, 255, 0x0340 ;,  0x2600, 128

	; Memory bank 3 : Matsushita Electric MCN-14EAEC (6V, 40p/r)
	.dw	300, 0x1E00, 0x0400, 0x0019, 230, 0x0840 ;,  0x1400, 64

	; Memory bank 4 :  
	.dw	300, 0x09C4, 0x0640, 0x0060, 180, 0x0550

	.org	$0100
;	.db	 (1<<TOIE0), (1<<INT0), DEF_SERVO_CYCLE, DEF_SERVO_CYCLE_SOFT, QUARTZ_FREQ
	; QUARTZ_FREQ -   ,   


.equ	N_PARM = 6;8	; Number of parameters par bank.
;----------------------------------------------------------;


.cseg
		.org	$0000

	RJMP 	Start     		;   reset
							;     
							; (- )    
	RJMP 	INT0_INT		; External Interrupt Request 0
	RJMP 	INT1_INT		; External Interrupt Request 1
	RJMP 	Start			; PCINT0		; Pin Change Interrupt Request 0
	RJMP 	Start			; PCINT1		; Pin Change Interrupt Request 0
	RJMP 	Start			; PCINT2		; Pin Change Interrupt Request 1
	RJMP 	Start			; WDT			; Watchdog Time-out Interrupt
	RJMP 	Start			; TIM2_COMPA	; Timer/Counter2 Compare Match A
	RJMP 	Start			; TIM2_COMPB	; Timer/Counter2 Compare Match B
	RJMP 	TIMER2_OVF_INT	; Timer/Counter2 Overflow
	RJMP 	Start			; TIM1_CAPT		; Timer/Counter1 Capture Event
	RJMP 	Start			; TIM1_COMPA	; Timer/Counter1 Compare Match A
	RJMP 	Start			; TIM1_COMPB	; Timer/Counter1 Compare Match B
	RJMP 	Start			; TIM1_OVF		; Timer/Counter1 Overflow
	RJMP 	Start			; TIM0_COMPA	; TimerCounter0 Compare Match A
	RJMP 	Start			; TIM0_COMPB	; TimerCounter0 Compare Match B
	RJMP 	TIMER0_OVF_INT	; TIM_OVF	 	; Timer/Couner0 Overflow
	RJMP 	Start			; SPI_SRC		; SPI Serial Transfer Complete
	RJMP 	ReceiveInt		; USART Rx Complete
	RJMP 	TransmitInt		; USART, Data Register Empty
	RJMP 	Start			; USART_TXC		; USART Tx Complete
	RJMP	Start			; ADC Conversion Complete
	RJMP 	Start			; EE_RDY		; EEPROM Ready
	RJMP 	Start			; ANA_COMP		; Analog Comparator
	RJMP 	Start			; TWI			; Two-wire Serial Interface
	RJMP 	Start			; SPM_RDY		; Store Program Memory Read

	
Start:
		clr		_0
		ldi		r16, low(RAMEND)
		out		spl, r16
		ldi		r16, HIGH(RAMEND)
		out		sph, r16
;-------------------------------------------------------
		wdr								;   
		lds 	r16,WDTCSR				;
		ori		r16,(1<<WDCE)|(1<<WDE)	;	
		sts 	WDTCSR,r16				;  WDT 
		ldi		r16,(1<<WDE)|(1<<WDP0)	;   
		sts 	WDTCSR,r16				; t=32ms
;-----------------------------------------------------------

		;  RAM
		ldi		YL, LOW(0x100)
		ldi		YH, HIGH(0x100)

		ldi		r16, 0 ;
		ldi		r17, 4 ; 1024 bytes
cl_ram:	st		Y+, _0
		sec
		sbc		r16, _0
		sbc		r17, _0
		cp		r16, _0
		cpc		r17, _0
		brne	cl_ram
		;-------

; ----------------------------(  (pin)  ATMega88   TQPF-32)
;**********   B: **********************************************
	ldi	r16,0b11111111			; PB7(pin8) - XTAL2
	out	DDRB,r16				; PB6(pin7) - XTAL1
	ldi	r16,0b00000000			; PB5(pin17) - SCK
	out	PORTB,r16				; PB4(pin16) - MISO
								; PB3(pin15) - MOSI
;     	; PB2(pin14) -  1 (-2)
;     			; PB1(pin13) -  1 (-1)
;  ""  0				; PB0(pin12) -  ( .)
	
;**********   : **********************************************
	ldi	r16,0b11110110			; PC7(xxx) -    
	out	DDRC,r16				; PC6(pin29) - Reset ( .)
	ldi	r16,0b00000000			; PC5(pin28) -  ( .) 
	out	PORTC,r16				; PC4(pin27) -  ( .)
								; PC3(pin26) -  (DIRECT (F/R) - ) 
							  	; PC2(pin25) -  ( .)
								; PC1(pin24) -  ( .)
								; PC0(pin23) -  ( .)
	
;**********   D: **********************************************
	ldi	r16,0b11001010			; PD7(pin11) -  ( .)
	out	DDRD,r16				; PD6(pin10) -  ( .)
	ldi	r16,0b00000000			; PD5(pin9) -    
	out	PORTD,r16				; PD4(pin2) -    
								; PD3(pin1) -  ( .)
;     	; PD2(pin32) -  Step 
;     			; PD1(pin31) -  TX 
; ""  0				; PD0(pin30) -  RX 

;----------------------------------------------------------------
		ldi	r16,0b10000000		;  .
		out	ACSR,r16			;
;----------------------------------------------------------------
; timer 0 -  , -
		ldi		r16,0b00000011; CK/64 prescaler	
		out		TCCR0B, r16	; (976)
;---------------------------------------------------------------
; timer1 -  , (Phase correct PWM, 8-bit, )
		ldi		r16,0b10100001	;
		sts		TCCR1A, r16		;

		ldi		r16,0b00000001	;   
		sts		TCCR1B, r16
		
		sts		OCR1AH, _0
		sts		OCR1AL, _0
		sts		OCR1BH, _0
		sts		OCR1BL, _0
;-------------------------------------------------------------------
; timer 2 - - (  )
		ldi		r16,0b00000111	; CK/1024 prescaler	
		sts		TCCR2B, r16
;--------------------------------------------------------------------
		ldi		r16,0b00000001	;   
		sts		TIMSK0,r16		;    Timer0 
		sts		TIMSK2,r16		; & Timer2
;---------------------------------------------------------------------
		ldi		r16,0b00000010	;INT0    (rising ISC01 | ISC00)
		sts		EICRA, r16		;
	
		ldi		r16,0b00000001	;   INT0 (Step)
		out		EIMSK, r16		;
;-----------------------------------------------------------------

	;  UART, 38400 bps
		sts		UBRR0H, _0
		ldi		r16,0b00011001
		sts		UBRR0L, r16

		ldi		r16, (1<<RXEN0)|(1<<TXEN0) ;|(1<<CHR9)|(1<<TXB8)
		ldi		r16,0b00011000
		sts		UCSR0B, r16

		ldi		r16,0b00000110 			; 8-bit/no parity/1 stop
		sts		UCSR0C, r16
;---------------------------------------------------------------------	
	   	sts 	UART_OUT_T,_0
	   	sts 	UART_OUT_H,_0

		clr		_Flags
		sbr		_Flags, (1<<wait_cmd_bit) ; 

		wdr							;   

		;  3- -
		ldi		r16, 3
		mov		_Mode, r16
		rcall	InitServo


		rcall LongDelay
		wdr								;   
		rcall LongDelay


		;   PID-   0 EEPROM
		clr		r16
		rcall	LoadPIDParamEEPROM

	
		ldi		r16,0b10011000
		sts		UCSR0B, r16

		sei						;   

		
		;   UART
		ldi		zl, LOW(HelloMsg*2)
		ldi		zh, HIGH(HelloMsg*2)
		rcall	WriteUARTString


Main:
		sbrc	_Flags, command_bit
		rcall	In_Message
		rjmp	Main


In_Message:
		lds		r16, UART_COMMAND

		cpi		r16, 'p' 			; Get Position
		breq	do_get_position

		cpi		r16, 'V' 			; Get Velocity and 1kHz counter
		breq	do_get_vel_1khz_cnt

		cpi		r16, 'B' 			; Set sub-command
		brne	PC+2
		rjmp	do_set_sub
			
		cpi		r16, 'P' 			; Set Position
		brne	PC+2
		rjmp	do_set_position
		
		cpi		r16, 'M' 			; Set Servo Mode
		brne	PC+2
		rjmp	do_set_servomode

		cpi		r16, 'm' 			; Get Servo Mode
		brne	PC+2
		rjmp	do_get_servomode

		cpi		r16, 'L' 			; Load from EEPROM bank
		brne	PC+2
		rjmp	do_load_eeprom_bank

		cpi		r16, 'S' 			; Save to EEPROM bank
		brne	PC+2
		rjmp	do_save_eeprom_bank
		
		cpi		r16, 'I' 			; Get Welcome string and Version
		brne	PC+2
		rjmp	do_welcome_str

		cpi		r16, 'W' 			; Write Param X
		brne	PC+2
		rjmp	do_write_param1

		cpi		r16, 'R' 			; Read param X
		brne	PC+2
		rjmp	do_read_param1

		cpi		r16, 'z' 			; Read position 2
		brne	PC+2
		rjmp	do_read_pozzz

		rjmp	cmd_fail			;  

; ++++++++++++++++++++++++++++++++++++
;  ! !
; ++++++++++++++++++++++++++++++++++++
do_get_position:
		cli
		mov		r16, _PosH
		mov		zl, _PosM
		mov		uart_r, _PosL
		sei
		rcall	MovsBuf
		mov		uart_r, zl
		rcall	MovsBuf
		mov		uart_r, r16
		rcall	MovsBuf
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  ! ,    1 kHz!
; ++++++++++++++++++++++++++++++++++++
do_get_vel_1khz_cnt:
		cli
		mov		r16, _200HzCntH
		mov		uart_r, _200HzCntL
		sei
		rcall	MovsBuf
		mov		uart_r, r16
		rcall	MovsBuf

		cli
		lds		r16, Velocity+1		;mov		r16, _VelocityH
		lds		uart_r, Velocity+0 	;mov		uart_r, _VelocityL
		sei	
		rcall	MovsBuf
		mov		uart_r, r16
		rcall	MovsBuf

		cli
		lds		r16, VelError200+1	;lds		r16, VelError+1
		lds		uart_r, VelError200+0;lds		uart_r, VelError+0
		sei	
		rcall	MovsBuf
		mov		uart_r, r16
		rcall	MovsBuf
		rjmp	cmd_ok


; ++++++++++++++++++++++++++++++++++++
;  ! SUB-COMMAND!
; ++++++++++++++++++++++++++++++++++++
do_set_sub:
		cli
		lds		r16, UART_IN+1
		sts		CtSub+0, r16
		lds		r16, UART_IN+0
		sts		CtSub+1, r16
		sei
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  ! !
; ++++++++++++++++++++++++++++++++++++
do_set_position:
		cli
		lds		r16, UART_IN+2
		sts		CtPos+0, r16
		lds		r16, UART_IN+1
		sts		CtPos+1, r16
		lds		r16, UART_IN+0
		sts		CtPos+2, r16
		sei
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !   EEPROM!
; ++++++++++++++++++++++++++++++++++++
do_load_eeprom_bank:
		lds		r16, UART_IN
		cpi		r16, (E2END+1)/N_PARM/2
		brcs    do_le
		rjmp	cmd_fail
do_le:	rcall	LoadPIDParamEEPROM
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !   EEPROM!
; ++++++++++++++++++++++++++++++++++++
do_save_eeprom_bank:
		lds		r16, UART_IN
		cpi		r16, (E2END+1)/N_PARM/2
		brcs    do_se
		rjmp	cmd_fail
do_se:	rcall	SavePIDParamEEPROM
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  !
; ++++++++++++++++++++++++++++++++++++
do_set_servomode:
		cli
		lds		_Mode, UART_IN+0
		rcall	InitServo
		sei
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  !
; ++++++++++++++++++++++++++++++++++++
do_get_servomode:
		mov		uart_r, _Mode
		rcall	MovsBuf
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  X!
; ++++++++++++++++++++++++++++++++++++
do_read_param1:
		lds		r16, UART_IN 			;  
		cpi		r16, N_PARM  			;   ?
		brcc    cmd_fail

		ldi		r17, 2
		clr		r19
srp_a:  subi	r16, 1
		brcs	srp_b
		add		r19, r17
		rjmp	srp_a

srp_b:  ldi		yl, low(Parms)
		ldi		yh, high(Parms)
		add		yl, r19
		adc		yh, _0

		ld 		uart_r, Y+
		rcall	MovsBuf
		ld 		uart_r, Y
		rcall	MovsBuf
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  X!
; ++++++++++++++++++++++++++++++++++++
do_write_param1:
		lds		r16, UART_IN+2 			;  
		cpi		r16, N_PARM 			;   ?
		brcc    cmd_fail

		ldi		r17, 2
		clr		r19
swp_a:  subi	r16, 1
		brcs	swp_b
		add		r19, r17
		rjmp	swp_a

swp_b:  ldi		yl, low(Parms)
		ldi		yh, high(Parms)
		add		yl, r19
		adc		yh, _0

		lds		r16, UART_IN+1
		st 		Y+, r16
		lds		r16, UART_IN
		st 		Y, r16
		rjmp	cmd_ok

do_welcome_str:
		ldi		zl, low(HelloMsg*2)
		ldi		zh, high(HelloMsg*2)
		rcall	WriteUARTString
		rjmp	cmd_ok

; ++++++++++++++++++++++++++++++++++++
;  !  2 (Z)!
; ++++++++++++++++++++++++++++++++++++
do_read_pozzz:
		cli
		lds		r16, CtPos+2
		lds		zl, CtPos+1
		lds		uart_r, CtPos+0
		sei
		rcall	MovsBuf
		mov		uart_r, zl
		rcall	MovsBuf
		mov		uart_r, r16
		rcall	MovsBuf
		rjmp	cmd_ok
;----------------------------------------------
cmd_fail:
		ldi		uart_r, '-'
		rjmp	PC+2
cmd_ok:	ldi		uart_r, '+'
		rcall	MovsBuf

		cbr		_Flags, (1<<command_bit) ;  
		sbr		_Flags, (1<<wait_cmd_bit);   

		ret


;/////////////////////////////////////////////////////////////////////////////////////////
;     UART_OUT   uart_r
MovsBuf: 	
        push 	zl                 		;     
        push 	zh
        push 	r16
        push 	r17

;		(cbi	UCSR0B, UDRIE0)  		;   UDRE(
        lds		r17,UCSR0B     			; ,     )
		andi	r17,0b11011111
		sts		UCSR0B,r17

        lds 	zl,UART_OUT_T      		;    
        inc 	zl              		;    1
        andi 	zl,$0f         			;     15   
										;  (   16 )

        lds 	r16,UART_OUT_H  		;    
        cp  	zl,r16          		;       
        brne 	m1                 		;      (   
	                                	; ),  m1

        sbr		_Flags, (1<<full_bit)	;    ,   
	                                 	;         
        rjmp 	endm               		;    


m1:     sts 	UART_OUT_T,zl     		;       
		cbr		_Flags, (1<<full_bit)	; C  ,   
                                 		;      
  ;      clr 	zh                 		;    - RAM
  		ldi		zh, high(UART_OUT)		;      RAM
        dec 	zl 
        andi 	zl,$0f
	;	subi	zl,UART_OUT       		; 
		ldi		r17, low(UART_OUT)     	;     UART_OUT  
		add		zl,r17					;   
   		st  	z,uart_r           		;    
endm:
		;(sbi		UCSRB, UDRIE)      	;   UDRE
		lds		r17,UCSR0B
		ori		r17,0b00100000
		sts		UCSR0B,r17

        pop 	r17
        pop 	r16		               	;    
        pop 	zh
        pop 	zl
        ret


;    UDRE
TransmitInt:
		push	zl						;    
		push	zh
		push	r16
		in		r16, SREG
		push	r16
		push	r17

		lds		zl, UART_OUT_H			;   
	;	clr		zh
		ldi		zh, high(UART_OUT)		;     RAM

;		(subi	zl,	-UART_OUT)			;      
  		ldi		r17, low(UART_OUT)     	;    UART_OUT  
		add		zl,r17					;  

		ld		r16, z					;    
		sts		UDR0, r16				;   UART

		lds		zl, UART_OUT_H		 	;     ?	
		inc		zl
		andi	zl, $0f
		lds		zh, UART_OUT_T
		cp		zl, zh
		brne	tl						;  ,   t1			

;		(cbi		UCSR0B, UDRIE0)	 	;  ,   	
		lds		r17,UCSR0B				;    UDRE ( )
		andi	r17,0b11011111
		sts		UCSR0B,r17		
						
tl:		sts		UART_OUT_H, zl			;     
                                 		;  .	
						
		pop		r17		
		pop		r16						;    
		out		SREG, r16
		pop		r16
		pop		zh
		pop		zl
		reti


;================================================================================
; UART RX INTERRUPT;
;          UART
;================================================================================
ReceiveInt:
		push	zl
		push	zh
		push	uart_r
		push	r16
		in		r16, SREG
		push	r16
		push	r17

		lds		r16, UDR0	

;		sbic	UCSR0A,DOR0
		lds		r17,UCSR0A
		sbrc	r17,DOR0				;  
		rjmp	qq	

		;DEBUG!!!!!!!!!!!!!!!!!!!!!!!!!1
		;mov		uart_r, r16
		;rcall	MovsBuf
		;rjmp	endr
		;DEBUG!!!!!!!!!!!!!!!!!!!!!!!!!1

		sbrs	_Flags, command_bit	
		rjmp	inl

qq:	
		sbr		_Flags, (1<<wait_cmd_bit)  ;   
		cbr		_Flags, (1<<wait_size_bit)

		ldi		uart_r, 'X'
		rcall	MovsBuf
		rjmp	endr

inl:	
		;  ?
		sbrs	_Flags, wait_cmd_bit
		rjmp	r_no_cmd	; 

		sts		UART_COMMAND, r16	;   (2 cyc)
		cbr		_Flags, (1<<wait_cmd_bit)
		sbr		_Flags, (1<<wait_size_bit) ;   
		rjmp	endr

r_no_cmd:
		;  -  
		sbrs	_Flags, wait_size_bit
		rjmp	r_no_siz	; 

		cpi		r16, 16		;  ?
		brlo	r_s_ok		; !
		rjmp	qq 			; 


r_s_ok:	
		tst		r16
		brne	r_st		
		sbr		_Flags, (1<<command_bit)
		rjmp	endr		


r_st:	dec		r16
		sts		UART_DATA_SIZE, r16 ;  
		cbr		_flags, (1<<wait_size_bit)
		rjmp	endr	

r_no_siz:
		;    
		lds		zl,	UART_DATA_SIZE
		mov		uart_r, zl
;		clr		zh
		ldi		zh, high(UART_IN)
 ;		subi	zl, -UART_IN
 		ldi		r17, low(UART_IN)     	;
		add		zl,r17			
	
		st		z, r16
		tst		uart_r	
		brne	r_dec	; , 
		sbr		_Flags, (1<<command_bit)	
		rjmp	endr

r_dec:	dec		uart_r
		sts		UART_DATA_SIZE, uart_r

endr:
		pop		r17	
		pop		r16
		out		SREG, r16
		pop		r16
		pop		uart_r
		pop		zh
		pop		zl
		reti

;================================================================================
;  
;================================================================================
InitServo:
		ldi		yl, low(CtPos)
		ldi		yh, high(CtPos)
		ldi		r16, 9
cl_pos:	st		Y+, _0
		dec		r16
		brne	cl_pos
		clr		_PosL
		clr		_PosM
		clr		_PosH
 		ret

;================================================================================
;   PID-  EEPROM
; : r16 -  
;================================================================================
SavePIDParamEEPROM:
		ldi		r17, N_PARM*2	; -  (word)
		clr		r19
smu_a:	subi	r16, 1			;   (r19 =   * - )
		brcs	se_co
		add		r19, r17		
		rjmp	smu_a
se_co:	ldi		yl, low(Parms)
		ldi		yh, high(Parms)
see:	sbic	EECR, EEPE
		rjmp	see				;   ( )
		out		EEARH, _0		;  AtMega48 -     ; (  )
		out		EEARL, r19		; EEPROM address (0..127)
		inc		r19
		ld		r16, Y+			; 
		out		EEDR, r16
		cli						; . , ..
		sbi		EECR, EEMPE		;   EEMWE   5 
		sbi		EECR, EEPE		;    EEWE
		sei
		dec		r17
		brne	see
		ret

;================================================================================
;   PID-  EEPROM
; : r16 -  
;================================================================================
LoadPIDParamEEPROM:
		ldi		r17, N_PARM*2 	; -  (word)
		clr		r19
lmu_a:	subi	r16, 1			;   (r19 =   * - )
		brcs	le_co
		add		r19, r17
		rjmp	lmu_a
le_co:	ldi		yl, low(Parms)
		ldi		yh, high(Parms)
wwco:	sbic	EECR, EEPE
		rjmp	wwco
lee:	
		out		EEARH, _0		;  AtMega48 -     ;(  )
		out		EEARL, r19		; EEPROM address (0..127)
		inc		r19
		sbi		EECR, EERE		; Read data (EEDR)
		in		r16, EEDR
		st		Y+, r16			; store data to RAM
		dec		r17
		brne	lee
		ret

;================================================================================
;   INT1 - 
;================================================================================
INT1_INT:
		push	r0
		in		r0, SREG

		sec

;		sbis	PIND, D_RESERVED
;		rjmp	i1_fall	

; rising edge
		sbis	PIND, D_ENCODER_B
		rjmp	i1sub
		rjmp	i1add
; falling edge
;i1_fall:
;		sbic	PIND, D_ENCODER_B
;		rjmp	i1sub

i1add:  cli
		adc		_PosL, _0
		adc		_PosM, _0
		adc		_PosH, _0
		sei
		rjmp	i1out

i1sub:  cli
		sbc		_PosL, _0
		sbc		_PosM, _0
		sbc		_PosH, _0
		sei

i1out:	out		SREG, r0
		pop		r0
		reti


;================================================================================
;   INT0 -    
;================================================================================
INT0_INT:
		push	r0
		in		r0, SREG		 ; Save carry flag
		push	zl

		sec
		lds		zl,	CtPos+0
		sbic	PINC, C_DIRECTION ; F/R = 0?	 
		rjmp	i0frw		      ; 
 
		; 
		sbc		zl, _0
		sts		CtPos+0, zl
		lds		zl, CtPos+1
		sbc		zl, _0
		sts		CtPos+1, zl
		lds		zl, CtPos+2
		sbc		zl, _0
		rjmp	i0so

i0frw:	; 
		adc		zl, _0
		sts		CtPos+0, zl
		lds		zl, CtPos+1
		adc		zl, _0
		sts		CtPos+1, zl
		lds		zl, CtPos+2
		adc		zl, _0

i0so:	sts		CtPos+2, zl
		pop		zl
		out		SREG, r0
		pop		r0
		reti


;================================================================================
;    TIMER2 - 16ms=62,5Hz
;================================================================================
TIMER2_OVF_INT:
		push	r0
		push	zl
		push	zh
		in		r0, SREG
		
;		lds		zl, ServoCycleTimeByte
		ldi		zl,0b00000000	;
		sts		TCNT2, zl		; clock divisor

		wdr							;   
		
		rjmp	tc2_Xkhz

tc2_eoi:
		out		SREG, r0
		pop		zh
		pop		zl
		pop		r0
		reti	


;================================================================================
;    TIMER0 - 83.3 
;================================================================================
TIMER0_OVF_INT:
		push	r0
		push	zl
		push	zh
		in		r0, SREG

		ldi		zl, -3 			; b11111101 (253, $FD)
		out		TCNT0, zl		; Restore Clock divisor


		mov		zl, _PvEnc		;   zl[1:0]
		
		in		_PvEnc, PIND	;     _PvEnc[5:4]
		swap	_PvEnc			;   _PvEnc[1:0]

		ldi		zh, 1

		sbrc	_PvEnc, 1		; Skip,    A  
		eor		_PvEnc, zh		; B xor 1

		; A B  A B
		; 1 1  1 0
		; 0 0  0 0
		; 1 0  1 1
		; 0 1  0 1

		sub		zl, _PvEnc		; Decode motion (zl = zl - _PvEnc)
		andi	zl, 3			; zl = zl and b11
		breq	enc_zr			; 00 -  
		cpi		zl, 1
		breq	enc_inc			; 01 -  
		cpi		zl, 3
		breq	enc_dec			; 11 -  

		; 10 - 
		mov		zl, _PvDir
		mov		zh, _PvDir		; Count by 2 for previous direction
		lsl		zl				; zl=zl shl 1, bit0 cleared
		asr		zh				; zh=zh shr 1, bit7=const, bit0 -> carry flag	
		rjmp	enc_add

enc_dec:
		ldi		zl, low(-1)
		ldi		zh, high(-1)
		rjmp	enc_setpvdir

enc_inc:
		ldi		zl, low(1)
		ldi		zh, high(1)
enc_setpvdir:
		mov		_PvDir, zl
enc_add:
		add		_PosL, zl		;  
		adc		_PosM, zh
		adc		_PosH, zh
enc_zr:
		dec		_CtDiv83		;  1/83 
		breq	tc0_1khz		;  0

tc0_eoi:
		out		SREG, r0
		pop		zh
		pop		zl
		pop		r0
		reti



;************************************************8
; SERVO, 1kHz
;************************************************8
tc0_1khz:
		ldi		zl, 83 ; (16MHz- 83, 110592 - 57.6, 10 MHz - 52; 8MHz - 42)
		mov		_CtDiv83, zl		; Re-initialize 1/83 divider

tc2_XkHz:
		sei
		push	r0
		push	r1
		push	r2
		push	r3
		push	r16
		push	r17
		push	r18
		push	r19
		push	r20
		push	r21
		push	r22

		;
		;
		; 200 HZ
		;
		dec		_CtDiv4			;  1/4 
		brne	no_200hz		;  0

		;   200Hz
		sec
		adc		_200HzCntL, _0
		adc		_200HzCntH, _0


		sts		Velocity+0, _VelocityL
		sts		Velocity+1, _VelocityH
		
		lds		zl, VelError+0
		sts		VelError200+0, zl
		lds		zl, VelError+1
		sts		VelError200+1, zl
		
		sts		VelError, _0
		sts		VelError+1, _0
		clr		_VelocityL
		clr		_VelocityH
		
		ldi		_CtDiv4, 5

no_200hz:
		lds		r2, PvPos+0		;  
		lds		r3, PvPos+1
		cli		
		sub		r2, _PosL
		sbc		r3, _PosM
		sts		PvPos+0, _PosL
		sts		PvPos+1, _PosM
		sei

		;   
		add		_VelocityL, r2
		adc		_VelocityH, r3

		com		r2				; $ff - r2
		com		r3				; $ff - r3
		adc		r2, _0
		adc		r3, _0			; r3:r2 = Velocity

		;  
		mov		r16, _Mode
		cpi		r16, 3			;  ( 3)
		breq	tap_position

		lds		r0, CtSub+0
		lds		r1, CtSub+1		; S- ( 0/1/2)
		mov		zl, r0
		mov		zh, r1
		mov		r19, r0
		mov		r20, r1
		cpi		r16, 2			;  ( 2)
		breq	tap_velocity
		cpi		r16, 1
		brne	PC+2
		rjmp	tap_torque		;   ( 1)
		rjmp	tap_voltage		;  ( 0)

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_position:
		cli
		lds		r16, CtPos+0	; 24-  
		lds		r17, CtPos+1
		lds		r18, CtPos+2
		
		sub		r16, _PosL		;   (dX)
		sbc		r17, _PosM
		sbc		r18, _PosH
		sei

		lds		r19, LimSpd+0   ;  
		lds		r20, LimSpd+1
		clr		r21
		cp		r16, r19		; 
		cpc		r17, r20
		cpc		r18, r21
		brge	pVovr; PC+11

		com		r19				; $ff - r19
		com		r20				; $ff - r20
		com		r21				; $ff - r21 = $ff

		adc		r19, _0
		adc		r20, _0
		adc		r21, _0

		cp		r16, r19
		cpc		r17, r20
		cpc		r18, r21
		brge	pVovr2 ; PC+3

pVovr:	mov		r16, r19
		mov		r17, r20
pVovr2:	mov		r0, r16
		mov		r1, r17			; r1:r0 = velocity command


;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
; r3:r2
; r1:r0 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_velocity:
		mov		r16, r2
		mov		r17, r3			; Velocity loop gain ( )

		lds		r18, GaSpd+0
		lds		r19, GaSpd+1
		rcall	muls1616		; r20:19 - scaled velocity

		sub		r0, r19			;  
		sbc		r1, r20

		;   -  
		lds		r16, VelError+0
		lds		r17, VelError+1
		add		r16, r0
		adc		r17, r1
		sts		VelError+0, r16	;  
		sts		VelError+1, r17	;  
		

		mov		r16, r0			;  
		mov		r17, r1			;  	
		lds		r18, GaTqP+0	;
		lds		r19, GaTqP+1	;
		rcall	muls1616		; r20:19 - proportional term
		mov		zl, r19
		mov		zh, r20
		
		lds		r16, PvInt+0	;  
		lds		r17, PvInt+1	;  	
		add		r0, r16			;
		adc		r1, r17			;
		mov		r16, r0			;
		mov		r17, r1			;
		lds		r18, GaTqI+0	;
		lds		r19, GaTqI+1	;
		rcall	muls1616		; r20:r19 - integral term
		add		zl, r19			;
		adc		zh, r20			; z:=proportional term + integral term

		lds		r18, LimTrq+0	;  
		ldi		r19, 0			;
		cp		zl, r18			;
		cpc		zh, r19			;
		brge	PC+6			;
		neg		r18				; $00 - r18
		com		r19				; $ff - r19
		cp		zl, r18			;
		cpc		zh, r19			;
		brge	PC+5			;
		mov		zl, r18			;
		mov		zh, r19			;
		sbr		_Flags, (1<<TorqLim_bit)	;
		rjmp	PC+6			;---
		sts		PvInt+0, r0		; Anti-wind-up
		sts		PvInt+1, r1		;
		cbr		_Flags, (1<<TorqLim_bit)	;


;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_torque:
		mov		r16, r2
		mov		r17, r3			; EG compensation
		lds		r18, GaEG+0		;
		lds		r19, GaEG+1		;
		rcall	muls1616		;
		add		r19, zl			;
		adc		r20, zh			; r20:r19 - Voltage command

;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
; 
; r20:r19
;$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
tap_voltage:
		ldi		r16, low(255)	;
		ldi		r17, high(255)	;   -255  +255.
		cp		r19, r16		;
		cpc		r20, r17		;
		brge	t6 ;PC+6		;

		ldi		r16, low(-255)	;
		ldi		r17, high(-255)	;
		cp		r19, r16		;
		cpc		r20, r17		;
		brge	mo ;PC+3		;

t6:		mov		r19, r16		;
		mov		r20, r17		;/ r20:r19


mo:		tst		r19				; ??
		brne	tdir			;
		cbi		PORTD, D_SWITCH1
		cbi		PORTC, C_SWITCH2
		sts		OCR1AH, _0
		sts		OCR1AL, _0
		sts		OCR1BH, _0
		sts		OCR1BL, _0
		rjmp	ep

tdir:
		tst		r20				;Motor output. Set PWM register and polarity
		brmi	sw1   	;PC+5	;
	
		; ONE DIR
		cbi		PORTC, C_SWITCH2
		cbi		PORTB, B_PWM_OUT2
		sts		OCR1BL, _0
		sts		OCR1AL, r19
		sbi		PORTD, D_SWITCH1
		rjmp	ep 		; PC+5	

sw1:
		; ANOTHER DIR
		cbi		PORTD, D_SWITCH1
		cbi		PORTB, B_PWM_OUT1
		neg		r19
		sts		OCR1AL, _0
		sts		OCR1BL, r19
		sbi		PORTC, C_SWITCH2

/*;---------------------------------------------------------------------------
mo: 	
	tst r19 						;   cbcb128 ( cnczone.com) 
	brne 	tdir 					;   Dead-Time  
	cbi 	PORTD, D_SWITCH1		; 
	cbi 	PORTC, C_SWITCH2
	cbr 	_Flags, (1<<pwma_bit) 	;Add to die area time

	cbr 	_Flags, (1<<pwmb_bit) 	;Close the designation of PWM 
	sts 	OCR1AL, _0
	sts 	OCR1AH, _0
	sts 	OCR1BL, _0
	sts 	OCR1BH, _0
	rjmp 	ep

tdir:
	tst 	r20 					;Motor output. Set PWM register and polarity
	brmi 	sw1 					;PC+5 ;

; ONE DIR
	cbi 	PORTC,C_SWITCH2
	cbi 	PORTB,B_PWM_OUT2
	sts 	OCR1BL, _0
	sbrs 	_Flags, pwma_bit 		;If the PWMA indicate to 1 jumps to turn
	rcall 	Delay1us 				;Delay1US 

	sts 	OCR1AL, r19
	sbi 	PORTD, D_SWITCH1
	sbr 	_Flags, (1<<pwma_bit) 	;??PWMA??
	cbr 	_Flags, (1<<pwmb_bit) 	;??PWMB??
	rjmp 	ep 						; PC+5 

sw1:
		; ANOTHER DIR
	cbi 	PORTD, D_SWITCH1
	cbi 	PORTB, B_PWM_OUT1
	neg 	r19
	sts 	OCR1AL, _0
	sbrs 	_Flags, pwmb_bit
	rcall 	Delay1us
	sts 	OCR1BL, r19
	sbi 	PORTC, C_SWITCH2
	sbr 	_Flags, (1<<pwmb_bit)
	cbr 	_Flags, (1<<pwma_bit)
;-----------------------------------------------------------------
*/

ep:		pop		r22
		pop		r21
		pop		r20
		pop		r19
		pop		r18
		pop		r17
		pop		r16
		pop		r3
		pop		r2
		pop		r1
		pop		r0
		sbr		_Flags, (1<<_1khz_bit)	; 1kHz interrupt
		rjmp	tc0_eoi




Delay1us:
;	nop
	rjmp	PC+1
	ret



; == END SERVO 1kHz ================



;================================================================================
;  16*16 (signed)
; : 32- (r21:r20:r19:r18)
; : 16- (r19:r18)
; : 16- (r17:r16)
;================================================================================
muls1616:
		clt
		tst		r17
		brpl	PC+6
		set
		com		r16
		com		r17
		adc		r16,_0
		adc		r17,_0

		sub		r20,r20	; clear high 16bit.
		sub		r21,r21	;
		ldi		r22,17	; r22 = loop count
		brcc	PC+3	; ---- calcurating loop
		add		r20,r16	;
		adc		r21,r17	;
		ror		r21	;
		ror		r20	;
		ror		r19	;
		ror		r18	;
		dec		r22	; if (--r22 > 0)
		brne	PC-8	;  continue loop;

		brtc	mulret ;PC+9	; Negate the result if multiplyer was negative.
		com		r18
		com		r19
		com		r20
		com		r21
		adc		r18, _0
		adc		r19, _0
		adc		r20, _0
		adc		r21, _0
mulret:	ret



;================================================================================
;    UART (Input - Z)
;================================================================================
WriteUARTString:
		lpm
		adiw	zl, 1
		tst		r0
		breq	wur
		mov		uart_r, r0
		rcall	MovsBuf
		rjmp	WriteUARTString
wur:	ret

;================================================================================
; 
;================================================================================
LongDelay:
		wdr							;   
		push	r16
		push 	r17

		ldi		r17, 255
		ldi		r16, 255

DELY1:	dec 	r17
		brne	DELY1
		dec 	r16
		brne 	DELY1

		pop 	r17
		pop		r16
		ret



HelloMsg:
		.db		"SERVO V0.2 INT!", 0
